/* sc-cron.c
 * SommerCampus 2006
 * Linux/Unix Systemprogrammierung
 *
 * Licence: GPL-2 */

#include "sc-cron.h"

static cron_entry *cronjobs_list=NULL;
static int execute_cron = 0;
static int reload_need = 0;

int daemonize(void)
{
	pid_t child;

	if((child = fork()) == -1)
	{
		syslog(LOG_ERR, "cannot create a child process");
		fprintf(stderr, "cannot create a child process");

		return 0;
	}

	if(child)
		exit(EXIT_SUCCESS); /* terminate parent process and let
				       child work as session leader */

	setsid();

	chdir("/");

	umask(0);
	
	return 1;
}

void synchronize_time(void)
{
	time_t unix_time;
	struct tm *su_time;

	while(1)
	{
		if(time(&unix_time) == -1)
			return;

		su_time = localtime(&unix_time);

		if(su_time->tm_sec == 0)
			return;
	}
}

void load_cronjobs(int signr)
{
	/* better solution: semaphors */
	if(execute_cron)
	{
		reload_need = 1;
		return;
	}

	free(cronjobs_list);

	cronjobs_list = get_all_cron_jobs();
	synchronize_time();
}

void cleanup(int signr)
{
	execute_cron = 1;
	if(unlink(PIDFILE))
		syslog(LOG_ERR, "Pid file %s can not be removed", PIDFILE);

	syslog(LOG_INFO, "server shutting down");

	closelog();

	free(cronjobs_list);
	exit(EXIT_SUCCESS);
}

int main(void)
{
	FILE* fp;
	size_t i;
	cron_time actual_time;
	
	if(getuid())
	{
		fprintf(stderr, "Only root can start sc-cron\n");
		return EXIT_FAILURE;
	}
	
	openlog("sc-cron", LOG_PID, LOG_CRON);
	syslog(LOG_INFO, "staring server");

	if(!daemonize())
	{
		fprintf(stderr, "sc-cron cannot be start as a daemon\n");
		syslog(LOG_ERR, "sc-cron cannot be start as a daemon");
		return EXIT_FAILURE;
	}


	synchronize_time();

	fp = fopen(PIDFILE, "w");

        if(!fp)
        {
		syslog(LOG_ALERT, "%s cannot be written: %m", PIDFILE);
                return EXIT_FAILURE;
        }

        fprintf(fp, "%d\n", getpid());

        fclose(fp);


	if(signal(SIGINT, cleanup) == SIG_ERR)
        {
                syslog(LOG_ALERT, "The signal SIGINT could not be set");
		free(cronjobs_list);
                return EXIT_FAILURE;
        }

	if(signal(SIGHUP, load_cronjobs) == SIG_ERR)
        {
                syslog(LOG_ALERT, "The signal SIGKILL could not be set");
		free(cronjobs_list);
                return EXIT_FAILURE;
        }
	
	cronjobs_list = get_all_cron_jobs();

	while(1)
	{
		execute_cron = 0;

		if(!get_actual_cron_time(&actual_time))
		{
			syslog(LOG_ALERT, "could not get the current time... aborting");
			return EXIT_FAILURE;
		}		

		/* execute cron jobs */
		execute_cron = 1;

		if(cronjobs_list)
		{
			/* there is at least on job to be executed */
			for(i = 0; i < cronjobs_list[0].count; ++i)
			{
				if(match_cron_time(&actual_time, &cronjobs_list[i].ctime))
					execute_cron_job(&cronjobs_list[i]);
			}
			
		}

		if(reload_need)
		{
			free(cronjobs_list);
			cronjobs_list = get_all_cron_jobs();
			reload_need = execute_cron = 0;
		}

		execute_cron = 0;
		sleep(60);
	}

	free(cronjobs_list);

	return EXIT_SUCCESS;
}

void execute_cron_job(const cron_entry *centry)
{
	pid_t avoid_zombie, child, exec_child;
	int status;
	struct passwd *uinfo;
	
	/* this function creates a child process, which parses the command
	 * and execute the command.
	 * this function avoids the builing of zombies, it creates a child 
	 * process in the first child process, so that becomes the child
	 * of init(8) */

	if((avoid_zombie = fork()) == -1)
	{
	
		syslog(LOG_ERR, "cannot execute cron job `%s` of user %s", centry->command, centry->user);
		return;
	}

	if(avoid_zombie)
	{
		/* parent process == cron server (DO NOT PUT exit(...) HERE), waits for child */
		waitpid(avoid_zombie, NULL, 0);
		return;
	} else {
		/* child process. Generates a new child, which will
		 * execute the desired code and close itselfs */

		if((child = fork()) == -1)
		{
			syslog(LOG_ERR, "cannot execute cron job `%s` of user %s", centry->command, centry->user);
			exit(EXIT_FAILURE);
		}

		if(child) /* parent process, exit and let child become a child of init(8) */
			exit(EXIT_SUCCESS);

		/* real child process (grandchild) which executes the cron jobs */

		if((exec_child = fork()) == -1)
		{
			syslog(LOG_ERR, "Could create a process for the cron job");
			exit(EXIT_FAILURE);
		}

		if(exec_child)
		{
			waitpid(exec_child, &status, 0);

			free(cronjobs_list);

			if(WIFEXITED(status))
				exit(EXIT_SUCCESS);
			else {
				syslog(LOG_ERR, "cron job terminated abnormaly\n");
				exit(EXIT_FAILURE);
			}
		} else {
			/* execute proccess */

			/* get user information and change (e)uid/(g)guid and
			 * HOME DIR */

			uinfo = getpwnam(centry->user);

			if(!uinfo)
			{
				syslog(LOG_ERR, "Cannot get information of user %s\n", centry->user);
				exit(EXIT_FAILURE);
			}

			syslog(LOG_INFO, "(%s) CMD (%s)", centry->user, centry->command);

			setuid(uinfo->pw_uid);
			seteuid(uinfo->pw_uid);
			setgid(uinfo->pw_gid);
			setegid(uinfo->pw_gid);

			if(setenv("HOME", uinfo->pw_dir, 1) == -1)
				syslog(LOG_ERR, "Cannot set environment variable HOME");

			chdir(uinfo->pw_dir);
			
			exit(execlp(SHELLPATH, SHELLPATH, "-c", centry->command, NULL));
		}

		exit(EXIT_SUCCESS);
	}
}
