/* sc-crontab.c
 * SommerCampus 2006
 * Linux/Unix Systemprogrammierung
 *
 * Licence: GPL-2 */

#include "sc-crontab.h"

void parse_input(int argc, char **argv, in_user *uvals)
{
	int i, j=0;
	struct passwd *uinfo;
	uvals->uid = uvals->euid = getuid();
	uvals->remove = 0;

	uinfo = getpwuid(uvals->uid);
	if(!uinfo)
	{
		fprintf(stderr, "could not get information of current user\n");
		exit(EXIT_FAILURE);
	}

	strcpy(uvals->uname, uinfo->pw_name);
	strcpy(uvals->euname, uinfo->pw_name);

	for(i=1; i < argc; ++i)
	{
		if(argv[i][0] == '-')
		{
			if(!strcmp(argv[i], "-r"))
				uvals->remove = 1;
			else if(!strcmp(argv[i], "--help"))
				help(argv[0], NULL, 0);
			else
				help(argv[0], "Unknown parameter", 1);
		} else {
			if(j++)
				help(argv[0], "You can edit/delete crontabs of one user only!", 1);

			if(uvals->uid && strcmp(argv[i], uvals->uname))
				help(argv[0], "You are not the super user", 1);

			if(!uvals->uid && strcmp(argv[i], "root"))
			{
				uinfo = getpwnam(argv[i]);

				if(!uinfo)
				{
					fprintf(stderr, "could not get information of user %s\n", argv[1]);
					exit(EXIT_FAILURE);
				}

				strcpy(uvals->euname, uinfo->pw_name);
				uvals->euid = uinfo->pw_uid;
			}
		}
	}

	sprintf(uvals->path[0], SPOOLER "/crontabs/%s", uvals->euname);
	sprintf(uvals->path[1], SPOOLER "/tmp/%s", uvals->euname);
}

void help(const char *prog, const char* error, int status)
{
	char fmt[] = "usage: %s [-r] [username] [--help]\n";
	
	if(status)
		fprintf(stderr, "%s\n", error);
	
	fprintf(status ? stderr : stdout, fmt, prog);

	exit(status);
}


/* read 1024 bytes and copy them into source */
int copy(const char *src, const char *dest, uid_t dest_uid, gid_t dest_gid, mode_t dest_mod)
{
	int fd[2]; /* 0: src, 1: dest */
	int block[1024];
	size_t read_bytes, write_bytes;

	fd[0] = open(src, O_RDONLY);

	if(!fd[0])
		return 0;

	fd[1] = creat(dest, dest_mod);

	if(!fd[1])
	{
		close(fd[0]);
		return 0;
	}

	if(fchown(fd[1], dest_uid, dest_gid))
	{
		fprintf(stderr, "The file rights for %s could not be set. Aborting...\n", dest);
		close(fd[1]);
		close(fd[0]);

		return 0;
	}

	while( (read_bytes = read(fd[0], block, 1024)) != -1 && read_bytes  )
	{
		write_bytes = write(fd[1], block, read_bytes);
		if(write_bytes == -1 || write_bytes != read_bytes)
			fprintf(stderr, "Unexpected error. The file %s may not be copied completely\n", dest);
	}

	close(fd[0]);
	close(fd[1]);

	return 1;
}

pid_t get_pid_from_cron_server(void)
{
	FILE* fp;
	int pid;

	fp = fopen(PIDFILE, "r");

	if(!fp)
	{
		if(errno == ENOENT)
			return 0; /* cron server may not be running */

		return -1; /* some badly error */
	}

	fscanf(fp, "%d", &pid);

	fclose(fp);

	return pid;
}

int main(int argc, char **argv)
{
	pid_t child, cron;
	char editor[1024], *tmp, 
	     message[] = "# crontab format\n# minute hour day month command\n\
# Use * instead of numbers, if you want that e.g. that all minutes on the minute field match\n\n";
			  
	int fd;
	
	in_user uinfo;

	parse_input(argc, argv, &uinfo);

	/* set default editor to vi */
	strcpy(editor, "vi");

	tmp = getenv("EDITOR");

	if(tmp)
		strcpy(editor, tmp);

	/* become root */
	if(setuid(0))
	{
		perror("setuid");
		return EXIT_FAILURE;
	}

	openlog("sc-crontab", LOG_PID, LOG_CRON);

	if(uinfo.remove)
	{
		fd = open(uinfo.path[0], O_RDONLY);

		if(fd == -1)
		{
			if(errno == ENOENT)
			{
				fprintf(stderr, "You don't have any cron jobs to remove\n");
				return EXIT_FAILURE;
			} else
				fprintf(stderr, "Your spooler may not be removed\n");
		} else
			close(fd);

		if(unlink(uinfo.path[0]))
		{
			fprintf(stderr, "Your cron jobs couldn't be removed\n");
			return EXIT_FAILURE;
		}

		syslog(LOG_INFO, "(%s) DELETE (%s)", uinfo.uname, uinfo.euname);

		printf("Cron jobs successfully removed.... ");

		cron = get_pid_from_cron_server();

		if(cron == -1)
			printf("could not get PID file of sc-cron. Please restart the cron server manually\n");
		else if(cron == 0)
			printf("sc-cron is not running and/or is not ready\n");
		else {
			if(kill(cron, SIGHUP))
				printf("sc-cron could not receive any signal: %s\nPlease restart the cron server manually\n", strerror(errno));
			else
				printf("successfull\n");
		}

		return EXIT_SUCCESS;
	}

	/* are there already cron jobs for current user? */

	fd = open(uinfo.path[0], O_RDONLY);

	if(fd == -1)
	{
		/* no there are not */
		if(errno != ENOENT)
		{
			fprintf(stderr, "cannot determine cron jobs for user %s\n", uinfo.euname);
			return EXIT_FAILURE;
		}

		fd = creat(uinfo.path[1], S_IRUSR | S_IWUSR); /* chmod 0600 */

		if(fd == -1)
		{
			fprintf(stderr, "cannot create a spooler file for cron jobs for user %s\n", uinfo.euname);
			return EXIT_FAILURE;
		}

		write(fd, message, strlen(message));

		close(fd);
	} else {
		close(fd);
		if(!copy(uinfo.path[0], uinfo.path[1], 0, 0, S_IRUSR | S_IWUSR))
		{
			fprintf(stderr, "a safe copy of %s cannot be generated. Aborting...\n", uinfo.path[0]);
			return EXIT_FAILURE;
		}
	}

	child = fork();

	if(child == -1)
	{
		perror("fork");
		return EXIT_FAILURE;
	}

	if(child)
	{
		/* parent proc */
		waitpid(child, &fd, 0);

		if(WIFEXITED(fd))
		{
			if(WEXITSTATUS(fd))
			{
				fprintf(stderr, 
					"The editor terminated with following error: %s\nThe cronjobs may not be saved\n", strerror(WEXITSTATUS(fd)));
				return EXIT_FAILURE;
			}

			if(!copy(uinfo.path[1], uinfo.path[0], uinfo.euid, 0, S_IRUSR | S_IWUSR))
			{
				fprintf(stderr, "The cron jobs could not be saved properly. The cronjobs may not be saved\n");
				return EXIT_FAILURE;
			}

			unlink(uinfo.path[1]);

			printf("Your cronjobs were successfully saved. Installing the cron jobs ... ");

			syslog(LOG_INFO, "(%s) REPLACE (%s)", uinfo.uname, uinfo.euname);

			cron = get_pid_from_cron_server();

			if(cron == -1)
				printf("could not get PID file of sc-cron. Please restart the cron server manually\n");
			else if(cron == 0)
				printf("sc-cron is not running and/or is not ready\n");
			else {
				if(kill(cron, SIGHUP))
					printf("sc-cron could not receive any signal: %s\nPlease restart the cron server manually\n", strerror(errno));
				else
					printf("successfull\n");
			}
			
			closelog();
			exit(EXIT_SUCCESS);

		} else {
			fprintf(stderr, "The editor didn't terminated normaly. The cronjobs may not be saved\n");
			return EXIT_FAILURE;
		}

		
	} else {
		/* child proc */
		execlp(editor, editor, uinfo.path[1], NULL);

		exit(errno);
	}

	closelog();

	return EXIT_SUCCESS;
}
